//============================================================================
// Name        : uhidled.cpp
// Author      : Aleph
// Email	   : aleph.at.al3ph.org
// Version     : 1.0.0
// Copyright   : Silicon Based Life (2009)
// Description : U-HID LED Controller
//============================================================================

/*
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#define HAVE_STDBOOL_H

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <argp.h>
#include <hid.h>
#include <memory.h>

const char *argp_program_version = PACKAGE_STRING;
const char *argp_program_bug_address = PACKAGE_BUGREPORT;

/* Program documentation. */
static char doc[] = "uhidled -- a program to query/write, U-HID LEDs";

/* A description of the arguments we accept. */
static char args_doc[] = "";

/* The options we understand. */
static struct argp_option options[] = {
	{"verbose",  	'v', 0,      		0,  "Produce verbose output" },
	{"quiet",    	'q', 0,      		0,  "Don't produce any output" },
	{"silent",   	's', 0,      		OPTION_ALIAS },
	{"pacdrive",	'p', 0, 			0,  "Pacdrive mode" },
	{"controller",	'c', "[1-8]", 		0,  "Select controller to modify" },
	{"all",			'a', "[0,1]", 		0,  "Set all LED`s to this state" },
	{"hex",			'x', "[0xFF]", 		0,  "Set LED`s using 16 bit hex value" },
	{"l",			'l', "[0,1,..]", 	0,  "Set LED`s comma separated list" },
	
	{ 0 }
};  

/* Used by main to communicate with parse_opt. */
struct arguments {
	int silent, verbose,pacdrive;
	char *all;
	int controller;
	char *hex;
	char *l;
};

/* Parse a single option. */
static error_t parse_opt (int key, char *arg, struct argp_state *state) {
	/* Get the input argument from argp_parse, which we
	 know is a pointer to our arguments structure. */
	struct arguments *arguments = (struct arguments *)state->input;

	switch (key) {
		case 'q': case 's':
			arguments->silent = 1;
			break;

		case 'v':
			arguments->verbose = 1;
			break;

		case 'p':
			arguments->pacdrive = 1;
			break;

		case 'x':
			arguments->hex = arg;
			break;

		case 'l':
			arguments->l = arg;
			break;

		case 'c':
			sscanf(arg,"%d",&(arguments->controller));
			break;
			
		case 'a':
			arguments->all = arg;
			if(!( 	strcmp(arguments->all,"0") == 0 || 
					strcmp(arguments->all,"1") == 0 || 
					strcmp(arguments->all,"-1") == 0 )){
				argp_usage (state);
			}
			break;

		default:
			return ARGP_ERR_UNKNOWN;
	}
	return 0;
}

/* argp parser. */
static struct argp argp = { options, parse_opt, args_doc, doc };

#define VENDOR_ID_ULTIMARC	0xD209
#define PRODUCT_ID_UHID1	0x1501

#define UHID_DATA_HEADER_LEN 2
#define UHID_DATA_DATA_LEN   2

typedef struct _UHID_DATA_BLOCK {
    char header[UHID_DATA_HEADER_LEN];
    char data[UHID_DATA_DATA_LEN];
} UHIDData;

static const char LED_BYTE_ALL_ON  = 0xff;
static const char LED_BYTE_ALL_OFF = 0x00;

#define UHID_PATHLEN 1
static int const UHID_PATH_IN[UHID_PATHLEN] =  { 0x00090001 };
static int const UHID_PATH_OUT[UHID_PATHLEN] = { 0x00010009 };

/*
 * Get LED state  (currently not supported by uhid)
 * Returns 0 on success 
 */
int get_led_state(HIDInterface *hid, UHIDData *data) {
   int result = -1;

   memset(data, 0, UHID_DATA_HEADER_LEN);
   result = hid_get_input_report(hid, UHID_PATH_OUT, UHID_PATHLEN, 
                                 (char *)data, sizeof(UHIDData));
   if (result == HID_RET_SUCCESS) {
       result = 0;
   } else {
       printf( "get_led_state: FAILED\n");
   }
#ifdef DEBUG
   unsigned int idx;
   for(idx=0; idx < sizeof(UHIDData); idx++) {
       printf("get_led_state: byte[%d] = 0x%x\n", idx, ((char *)data)[idx]);
   }
#endif

   return result; 
}

/* 
 * Set LED state
 * Returns 0 on success 
 */
int set_led_state(HIDInterface *hid, UHIDData *data) {
   int result = -1;

   memset(data, 0, UHID_DATA_HEADER_LEN);
#ifdef DEBUG
   unsigned int idx;
   for(idx=0; idx<4; idx++) {
       printf("set: byte %d: 0x%x\n", idx, ((char *)data)[idx] & 0xff);
   }
#endif
   result = hid_set_output_report(hid, UHID_PATH_IN, UHID_PATHLEN, 
                                  (char *)data, sizeof(UHIDData));
   if (result == HID_RET_SUCCESS) {
       result = 0;
   } else {
       printf("Error result of sending output report: %d\n", result);
   }
   return result; 
}


int main(int argc, char **argv) {
	HIDInterface* hid; 
	hid_return ret;
	
	struct arguments arguments;

    /* Default values. */
	arguments.silent = 0;
	arguments.verbose = 0;
	arguments.all = NULL;
	arguments.l = NULL;
	arguments.hex = NULL;
	arguments.controller = 1;

	arguments.pacdrive = 0;

	/* Parse our arguments; every option seen by parse_opt will
	   be reflected in arguments. */
	argp_parse (&argp, argc, argv, 0, 0, &arguments);

	UHIDData data;
	memset(&data, 0, sizeof(UHIDData));
	
	if(arguments.all){
		if(strcmp(arguments.all,"0") == 0){
			if(arguments.verbose)fprintf(stdout, "Setting all LED`s off\n");
			data.data[0] = LED_BYTE_ALL_OFF;
			data.data[1] = LED_BYTE_ALL_OFF;
		} else if(strcmp(arguments.all,"1") == 0){
			if(arguments.verbose)fprintf(stdout, "Setting all LED`s on\n");
			data.data[0] = LED_BYTE_ALL_ON;
			data.data[1] = LED_BYTE_ALL_ON;
		} else if(strcmp(arguments.all,"-1") == 0){
			if(arguments.verbose)fprintf(stdout, "Toggling all LED`s \n");
			data.data[0] = ~data.data[0];
			data.data[1] = ~data.data[1];
		}
	}

	if(arguments.hex){
		int unsigned val;
		sscanf(arguments.hex,"%x",&val);
		data.data[1] = (val & 0xFF);
		data.data[0] = (val>>8 & 0xFF);
	}
	
	if(arguments.l){
		short b =0;
		int i = 0;
		char *p = strtok(arguments.l,",");
		while(p){
			
			int v;
			if(sscanf(p,"%d",&v)){
				b |= v<<i;
			}
			i++;
			p = strtok(NULL,",");
		}
		data.data[1] = ((char*)&b)[0];
		data.data[0] = ((char*)&b)[1];
	}
	
#ifdef DEBUG
   unsigned int idx;
   for(idx=0; idx<4; idx++) {
       printf("set: byte %d: 0x%x\n", idx, ((char *)&data)[idx] & 0xff);
   }
#endif
	HIDInterfaceMatcher matcher = { VENDOR_ID_ULTIMARC, PRODUCT_ID_UHID1+arguments.controller-1-arguments.pacdrive, NULL, NULL, 0 };

	hid_set_debug(HID_DEBUG_ERRORS);
//	hid_set_debug(HID_DEBUG_ALL);
	hid_set_debug_stream(stderr);
	hid_set_usb_debug(0);

	ret = hid_init();
	if (ret != HID_RET_SUCCESS) {
	  	fprintf(stderr, "hid_init failed with return code %d\n", ret);
	  	return 1;
	}

	hid = hid_new_HIDInterface();
	if (hid == 0) {
	  	fprintf(stderr, "hid_new_HIDInterface() failed, out of memory?\n");
	  	return 1;
	}
	
	ret = hid_force_open(hid, 0, &matcher, 3);
	if (ret != HID_RET_SUCCESS) {
		fprintf(stderr, "hid_force_open failed with return code %d\n", ret);
		return 1;
	}
	
	if(arguments.verbose) {
		ret = hid_write_identification(stdout, hid);
		if (ret != HID_RET_SUCCESS) {
			fprintf(stderr, "hid_write_identification failed with return code %d\n", ret);
			return 1;
		}
	}
	
	if(set_led_state(hid, &data)){
		return 1;
	}
		
	ret = hid_close(hid);
	if (ret != HID_RET_SUCCESS) {
		fprintf(stderr, "hid_close failed with return code %d\n", ret);
		return 1;
	}
	
	hid_delete_HIDInterface(&hid);
	
	ret = hid_cleanup();
	if (ret != HID_RET_SUCCESS) {
		fprintf(stderr, "hid_cleanup failed with return code %d\n", ret);
		return 1;
	}
	
	return EXIT_SUCCESS;
}




